/* mc.c     91/03/26   intermediate code compiler and decompiler              */

#include "mini.h"

#define MAXVARS     256		/* maximum of local variables in a clause */
#define MAXGOALS    256		/* maximum number of goals of an prolog-program */
#define MAXTMPS     16		/* maximum of temporal variables in a clause */

#define VOID_VAR    0		
#define TEMP_VAR    1
#define LOCAL_VAR   2

#define Var(cp)     loc_vars[-3-(SopFetch(cp) / (SIZEMW / SIZECHAR))]	/* value of cp. local variable */
#define ShortVar(o) loc_vars[o]											/* value of o. local variable */
#define Tmp(cp)     tmp_vars[SopFetch(cp) / (SIZEMW / SIZECHAR)]		/* value of cp. temporal variable */
#define ShortTmp(o) tmp_vars[o]											/* value of cp. temporal variable */
#define Offset(o)   ((-3-(o))*((int)SIZEMW / (int)SIZECHAR))			/* address of offset in the actual stack frame */

/* checks if enough space is available for the code area and the copy/environment stack */
#ifdef stackchange
#define CodeStackCheck  if(codeptr>=(CODEPTR)copyptr) \					
	                    longjmp(compiler_err, OutOfMemoryError)
#else
#define CodeStackCheck  if(codeptr>=(CODEPTR)stackptr) \
	                    longjmp(compiler_err, OutOfMemoryError)
#endif

typedef struct {
	MWPTR       address;        /* address of free variable                 */
	SHORT       type;           /* void or temporary or local               */
	SHORT       offset;         /* offset for local variables               */
	SHORT       var_count;      /* occurence count of variable in a clause  */
	SHORT       var_nr;         /* current occurene number of variable      */
	SHORT       first_goal;     /* goal number of first occurence, head = 0 */
	SHORT       last_goal;      /* goal number of last occurence            */
	SHORT       tempadress;     /* address of temp-variable right-order writing */
} VAR_STRUCT;

typedef struct {
	SHORT       goal_count;     /* count of goals in a clause               */
	SHORT       headgoal_count; /* count of head temp goals in a clause     */
	SHORT       goal_nr;        /* currenct number of goal                  */
	SHORT       all_vars;       /* count of all variables                   */
	SHORT       head_vars;      /* count of head variables                  */
	SHORT       frame_size;     /* (count of local variables + 2) * SIZEMW  */
	VAR_STRUCT  var[MAXVARS];
} VAR_TAB;


static VAR_TAB  v;                      /* variable table */
MW copyspace_required_goal[MAXGOALS];   /* counts the copy space which is required for each goal from the current clause */
MW copyspace_required_head;				/* counts the copy space which is required for the head of the current clause */
int count_unify_lists = 0;				/* counts the recursive number of lists in a list */
short varoffset = 0;					/* current offset-position of a local variable in the current stack frame */
int lastgoal;							/* when the current goal is the last in the clause "lastgoal"=1 else "lastgoal"=0 */
int compile_V1_head;					/* true if VAM1P-Code is generated */
int flag_v1_xgoal_generated;			/* true if VAM1P-Code is generated */
MW copyspace_required = 0L;				/* checks how much space is required for the copystack for each clause */
CODEPTR headbeginptr, headendptr, goalbeginptr, goalendptr, v1_codeptr[MAXGOALS];	/* current compilation-code pointer adresses */
MWPTR   v1_procptr;						/* points to the procedure block of the current lastcall */
CODEPTR v1_headptr;						/* points to the code of the headcode of the selfrecursion */
static int current_goal = 0;			/* number of current goal */
static MW       loc_vars[MAXVARS];      /* local variable store for decompile   */
static MW       tmp_vars[MAXTMPS];      /* temp variable store for decompile    */
static MW       clause_name;            /* name of current compiled clause      */
static MW       clause_arity;           /* arity of current compiled clause     */
static int      first_goal;             /* flag for decompile                   */
static int      is_block;               /* block instruction                    */
static CODEPTR  codeptr;                /* pointer to the last generated code   */
static jmp_buf  compiler_err;           /* error recovery                       */

static void compile_goal(MWPTR goal);	
static void compile_struct(MW arity, MWPTR sptr);
static int  find_var(MWPTR var_address);
static void compute_var_offsets(void);
static void gen_head_code(MWPTR goal);
static void gen_head_struct(MW arity, MWPTR sptr);
static void gen_goal_code(MWPTR goal);
static void gen_goal_struct(MW arity, MWPTR sptr);

static void gen_nop_code(int instr);
static void gen_sop_code(int instr, int arg1);
static void gen_lop_code(int instr, MW arg1);
static void gen_sopsop_code(int instr, int arg1, int arg2);
static void gen_soploplop_code(int instr, int arg1, MW arg2, MW arg3);
static void gen_soploploplop_code(int instr, int arg1, MW arg2, MW arg3, MW arg4);
static void gen_soplop_code(int instr, int arg1, MW arg2);
static void gen_loplop_code(int instr, MW arg1, MW arg2);
static void gen_loploplopsop_code(int instr, MW arg1, MW arg2, MW arg3, int arg4);
static void gen_soploploplopsop_code(int instr, int arg1, MW arg2, MW arg3, MW arg4, int arg5);

static int goal_decompile(MWPTR gptr);
static int h_struct_decompile(MWPTR gptr, MW arity);
static int g_struct_decompile(MWPTR gptr, MW arity);

#ifdef zwischencodeausgabe
	extern FILE *zfile;		/* is needed for the immediate code output in a file */
#endif
MWPTR lastgptr;				/* pointer to goal      */

/* compile clause goal at begin or end, return code in headcodeptr       */
/* sets v, clause_name, clause_arity, is_block, codeptr, lastgptr, headbeginptr, goalbeginptr and compiler_err */

int compile_clause(MW goal, int atbegin, CODEPTR *headcodeptr)
{
MWPTR gptr,					/* pointer to goal      */
	head,					/* pointer to head      */
	sptr,					/* pointer to structure */
	pptr;					/* pointer to procedure */
short i;

ErrorNumber = setjmp(compiler_err);
if (ErrorNumber)
	{
	return(ErrorNumber);
	}
v.goal_count = 0;
v.headgoal_count = 1;
v.goal_nr = 0;
v.all_vars = 0;
v.head_vars = 0;
is_block = FALSE;
lastgoal = FALSE; 
count_unify_lists = 0;
gptr = &goal;
for (i=0;i<MAXVARS;i++)
   v.var[i].tempadress = -2; 
copyspace_required = 0L; 
StructDeref(goal, gptr);
if (Struct(goal) && (XtrFunctor(sptr = XtrStruct(goal)) == RuleFunctor))
	{
	/* compile clause */
	gptr = FirstArgPtr(sptr);
	compile_goal(head = gptr++);
	goal = *gptr;
	ValDeref(goal);
	/* compile body */
	while (Struct(goal) && (XtrFunctor(sptr = XtrStruct(goal)) == CommaFunctor))
		{
		gptr = FirstArgPtr(sptr);
		compile_goal(gptr++);
		goal = *gptr;
		ValDeref(goal);
		}
	compile_goal(gptr);
	compute_var_offsets();
	/* generate code */
	*headcodeptr = codeptr = AddClauseHeader(heapptr);
#ifdef Compiler_VAM1P_activate
	headbeginptr = codeptr;
	compile_V1_head = TRUE;
#endif
#ifdef space
	write_header_copyspace_required(heapptr) = &copyspace_required_head;
#endif
	gen_head_code(head);
	goal = *(gptr = head + 1);
	ValDeref(goal);
	while (Struct(goal) && (XtrFunctor(sptr = XtrStruct(goal)) == CommaFunctor))
		{
		gptr = FirstArgPtr(sptr);
		gen_goal_code(gptr++);
		goal = *gptr;
		ValDeref(goal);
		}
	lastgoal = TRUE; 
#ifdef Compiler_VAM1P_activate
	lastgptr = gptr;
	goalbeginptr = codeptr;
	compile_V1_head = FALSE;
#endif
	gen_goal_code(gptr);   /* lastcall */
#ifdef Compiler_VAM1P_activate
	mi_compiler(codeptr, write_commands);
#endif
	lastgoal = FALSE;  
	}
else
	{
	/* compile fact */
	compile_goal(gptr);
	compute_var_offsets();
	*headcodeptr = codeptr = AddClauseHeader(heapptr);
#ifdef space
	write_header_copyspace_required(heapptr) = &copyspace_required_head;
#endif

	gen_head_code(gptr);
	}
heapptr = AdjustCodePtr(codeptr);
codeptr = *headcodeptr;
FrameSize(codeptr) = v.frame_size;
pptr = find_proc(find_func(clause_name, clause_arity));
insert_clause(pptr, codeptr, atbegin);
indexing(pptr+1);
return(Success);
}


/* compile the goal at gptr            */
/* uses v, is_block and compiler_err  */

static void compile_goal(MWPTR gptr)
{
MW      goal;
MW      i;              /* var index or arity */
MWPTR   functor;        /* functor of goal    */
static long j = 1;      /* count goals for trace */

goal = *gptr;
StructDeref(goal, gptr);
if (RefUndef(goal))
	{
	i = find_var(gptr);
	if ((v.goal_count == 0) || (v.var[i].var_count++ == 0))
		longjmp(compiler_err, InvalidGoalError);
	v.var[i].last_goal = v.goal_count++;
	return;
	}
if (Atom(goal))
	{
	goal = XtrAtomIndex(goal);
	i = 0;
	if ((functor = find_func(goal, i)) == 0)
		longjmp(compiler_err, ErrorNumber);
	}
else if(Struct(goal))
	{
	gptr = XtrStruct(goal);
	functor = XtrFunctor(gptr);
	i = FunctArity(functor);
	goal = FunctIndex(functor);
	gptr = FirstArgPtr(gptr);
	compile_struct(i, gptr);
	}
else
	longjmp(compiler_err, InvalidGoalError);
if ((functor == CutFunctor) || (find_bip(goal, i) >= 0))
	{
	if (v.goal_count == 0)
		longjmp(compiler_err, InvalidGoalError);
	if (v.goal_count == v.headgoal_count)
		v.headgoal_count++;
	if (functor == BlockFunctor)
		{
		is_block = TRUE;
		v.goal_count++;
		}
	else if (functor == ExitBlockFunctor)
		v.goal_count++;
	}
else if (find_proc(functor) == 0)
	longjmp(compiler_err, OutOfMemoryError);
v.goal_count++;
}


/* compile struct sptr with arity arity  */

static void compile_struct(MW arity, MWPTR sptr)
{
MW      m;              /* tagged value cell */
MWPTR   mptr;           /* pointer to m      */
int     i;              /* var index         */

while (--arity >= 0)
	{
	m = *(mptr = sptr++);
	StructDeref(m, mptr);
	if (RefUndef(m))
		{
		i = find_var(mptr);
		v.var[i].var_count++;
		v.var[i].last_goal = v.goal_count;
		}
	else if (List(m))
		compile_struct(2, XtrList(m));
	else if (Struct(m))
		{
		mptr = XtrStruct(m);
		compile_struct(XtrArity(mptr), FirstArgPtr(mptr));
		}
	} 
}

/* finds local variable var_address in the var-table v */
static int find_var(MWPTR var_address)
{
int i;

i = v.all_vars;
while (--i >= 0)
	if (var_address == v.var[i].address)
		return(i);
if (v.all_vars == MAXVARS)
	longjmp(compiler_err, OutOfVarsError);
i = v.all_vars++;
v.var[i].address = var_address;
v.var[i].type = VOID_VAR;
v.var[i].var_count = 0;
v.var[i].var_nr = 0;
v.var[i].first_goal = v.goal_count;
return(i);
}


/* computes the offsets of the local variables relative to the current stack frame */
static void compute_var_offsets(void)
{
int   i;
SHORT o;
if (v.headgoal_count == v.goal_count)
	v.frame_size = 0;
else
	v.frame_size = 2 * (SIZEMW / SIZECHAR);
if (is_block)
	v.frame_size += BlockSize * (SIZEMW / SIZECHAR);
i = v.all_vars;
while (--i >= 0)
	if (v.var[i].var_count > 1)
		if (v.var[i].last_goal < v.headgoal_count)
			v.var[i].type = TEMP_VAR;
		else
			{
			v.var[i].type = LOCAL_VAR;
			v.frame_size += (SIZEMW / SIZECHAR);
			v.var[i].offset = -v.frame_size;
			if (v.var[i].first_goal < v.headgoal_count)
				v.head_vars++;
			}
o = 0;
for (i = 0; i < v.all_vars; i++)
	if (v.var[i].type == TEMP_VAR)
		if (o < MAXTMPS) 
			{
			v.var[i].offset = o * (SIZEMW / SIZECHAR);
			o++;
			}
		else
			{
			v.var[i].type = LOCAL_VAR;
			v.frame_size += (SIZEMW / SIZECHAR);
			v.var[i].offset = -v.frame_size;
			v.head_vars++;  /* if (v.var[i].first_goal < v.headgoal_count) */
			}
}


/* generates the intermediate code for a head of a clause at the adress gptr */
static void gen_head_code(MWPTR gptr)
{
MW      goal;

copyspace_required = 0L;
goal = *gptr;
StructDeref(goal, gptr);
CodeStackCheck;
if (Atom(goal))
	{
	clause_name = XtrAtomIndex(goal);
	clause_arity = 0;
	}
else /* Struct(goal) */
	{
	gptr = XtrStruct(goal);
	clause_name = FunctIndex(XtrFunctor(gptr));
	clause_arity = XtrArity(gptr);
	}
if (clause_arity > 0)
	gen_head_struct(clause_arity, FirstArgPtr(gptr));
if (++v.goal_nr == v.goal_count)
	gen_nop_code(c_nogoal);

copyspace_required_head = copyspace_required;
}


/* generates the intermediate-code for a structure sptr with arity arity */
static void gen_head_struct(MW arity, MWPTR sptr)
{
MW      m;
MWPTR   mptr, functor;
int     i;

while (--arity >= 0)
	{
	m = *(mptr = sptr++);
	StructDeref(m, mptr);
	if (RefUndef(m))
		{
		i = find_var(mptr);
		switch(v.var[i].type)
			{
			case VOID_VAR:
				copyspace_required++; /* copyspace required for a void_variable */
				TraceCopySpace_required;
				gen_nop_code(h_void);
				break;
			case TEMP_VAR:
				if (v.var[i].var_nr++ == 0)
					{
					copyspace_required++; /* copyspace required for a first_temp_variable */
					TraceCopySpace_required;
					if (v.var[i].offset == 0)
						gen_nop_code(h_fsttmp0);
					else
						gen_sop_code(h_fsttemp, (int) v.var[i].offset);
					}
				else
					if (v.var[i].offset == 0)
						gen_nop_code(h_nxttmp0);
					else
						gen_sop_code(h_nxttemp, (int) v.var[i].offset);
				break;
			case LOCAL_VAR:
				if (v.var[i].var_nr++ == 0)
					{
					copyspace_required++; /* copyspace required for a first_local_variable */
					TraceCopySpace_required;
					v.var[i].tempadress = -1;
					gen_sop_code(h_fstvar, (int) v.var[i].offset);
					}
				else
					if (v.var[i].tempadress == -1)
						gen_sop_code(h_nxtvar, (int) v.var[i].offset);
					else if (v.var[i].tempadress == -0)
						{
#ifdef variables_right_order_writing
						gen_sop_code(h_fsttemp, (int) v.var[i].tempadress); 
#else
						gen_sop_code(h_nxttemp, (int) v.var[i].tempadress);
#endif

						v.var[i].tempadress = v.var[i].offset;
						}
					else
					{
#ifdef variables_right_order_writing
						gen_sop_code(h_nxtvar, (int) v.var[i].offset);
#else
						gen_sop_code(h_nxtvar, (int) v.var[i].offset);
#endif
					}
				break;
			} /* switch */
		} /* if */
	else if (Nil(m))
		gen_nop_code(h_nil);
	else if(Int(m) || Atom(m))
		gen_lop_code(h_const, m);
	else if (List(m))
		{
		gen_nop_code(h_list);
		copyspace_required += 2; /* copyspace required for a list */
		TraceCopySpace_required;
		gen_head_struct(2, XtrList(m));
		}
	else /* Struct(m) */
		{
		mptr = XtrStruct(m);
		functor = XtrFunctor(mptr);
		gen_lop_code(h_struct, (MW) functor);
		copyspace_required += FunctArity(functor); /* copyspace required for a structure */
		TraceCopySpace_required;
		gen_head_struct(FunctArity(functor), FirstArgPtr(mptr));
		}
	} 
}


/* generates the intermediate code for a goal of a clause at the adress gptr */
static void gen_goal_code(MWPTR gptr)
{
MW      goal, name, arity, m;
MWPTR   pptr, mptr, functor;
int     i, inst;
int     bipnr;
static j=1;
copyspace_required = 0L;
flag_v1_xgoal_generated = FALSE;
TraceCountGoals(j);
j++;
if ((v.goal_nr >= MAXGOALS-1) || (current_goal >= MAXGOALS-1))
	{
	longjmp(compiler_err, OutOfVarsError);
	}

#ifdef Compiler_VAM1P_activate
	lastgptr = gptr;
#endif
goal = *gptr;
StructDeref(goal, gptr);
if (RefUndef(goal))
	{
	i = find_var(gptr);
	v.var[i].var_nr++;
	if (v.goal_nr == v.headgoal_count)
		gen_sopsop_code(c_metacall, (int) v.head_vars, g_meta);
	else
		gen_sop_code(c_metacall, g_meta);
	gen_sop_code(c_metacall, (int) v.var[i].offset);
	bipnr = -1;
	}
else
	{
	if (Atom(goal))
		{
		name = XtrAtomIndex(goal);
		arity = 0;
		functor = find_func(name, arity);
		}
	else /* Struct(goal) */
		{
		gptr = XtrStruct(goal);
		functor = XtrFunctor(gptr);
		name = FunctIndex(functor);
		arity = FunctArity(functor);
		gptr = FirstArgPtr(gptr);
		}
	if (functor == CutFunctor)
		{
		bipnr = 0;
		gen_nop_code(c_cut);
		if (v.goal_nr > v.headgoal_count)
			gen_nop_code(v.frame_size);
		}
	else if ((bipnr = find_bip(name, arity)) >= 0)
#ifdef space
			gen_soplop_code(c_builtin, (int) bipnr, (MW) &copyspace_required_goal[v.goal_nr]); 
#else
			gen_sop_code(c_builtin, (int) bipnr); 
#endif
	else
		{
		pptr = FunctProc(functor);
		inst = c_goal;
		if (arity > 0)
			{
			m = *(mptr = gptr);
			StructDeref(m, mptr);
			if (RefUndef(m))
				{
				i = find_var(mptr);
				if ((v.var[i].type == LOCAL_VAR) && (v.var[i].var_nr != 0))
					{
					gptr++;
					arity--;
					v.var[i].var_nr++;
					i = varoffset = v.var[i].offset; 
					inst = c_xgoal;
					}
				}
			}


/* -------------------------------------------------*/
/* generate V1_xgoal-command */
		flag_v1_xgoal_generated = FALSE;
#ifdef Compiler_VAM1P_activate
		if ((inst == c_xgoal) && (lastgoal == TRUE))
			mi_compiler(codeptr, write_xgoal); 	
#endif
		if (flag_v1_xgoal_generated == FALSE)
		{
/* -------------------------------------------------*/

		if (v.goal_nr == v.headgoal_count)
		{
			switch (v.head_vars)
				{
				case 1:
					if (inst == c_goal)
#ifdef space
						gen_loplop_code(c_1_goal, (MW) pptr, (MW) &copyspace_required_goal[v.goal_nr]);
#else
						gen_lop_code(c_1_goal, (MW) pptr);
#endif
					else
#ifdef space
						gen_loplop_code(c_1_xgoal, (MW) pptr, (MW) &copyspace_required_goal[v.goal_nr]);
#else
						gen_lop_code(c_1_xgoal, (MW) pptr);
#endif
					break;
				case 2:
					if (inst == c_goal)
#ifdef space
						gen_loplop_code(c_2_goal, (MW) pptr, (MW) &copyspace_required_goal[v.goal_nr]);
#else
						gen_lop_code(c_2_goal, (MW) pptr);
#endif
					else
#ifdef space
						gen_loplop_code(c_2_xgoal, (MW) pptr, (MW) &copyspace_required_goal[v.goal_nr]);
#else
						gen_lop_code(c_2_xgoal, (MW) pptr);
#endif
					break;
				case 3:
					if (inst == c_goal)
#ifdef space
						gen_loplop_code(c_3_goal, (MW) pptr, (MW) &copyspace_required_goal[v.goal_nr]);
#else
						gen_lop_code(c_3_goal, (MW) pptr);
#endif
					else
#ifdef space
						gen_loplop_code(c_3_xgoal, (MW) pptr, (MW) &copyspace_required_goal[v.goal_nr]);
#else
						gen_lop_code(c_3_xgoal, (MW) pptr);
#endif
					break;
				case 4:
					if (inst == c_goal)
#ifdef space
						gen_loplop_code(c_4_goal, (MW) pptr, (MW) &copyspace_required_goal[v.goal_nr]);
#else
						gen_lop_code(c_4_goal, (MW) pptr);
#endif
					else
#ifdef space
						gen_loplop_code(c_4_xgoal, (MW) pptr, (MW) &copyspace_required_goal[v.goal_nr]);
#else
						gen_lop_code(c_4_xgoal, (MW) pptr);
#endif
					break;
				default:
					TraceCodeGeneration("default-Zweig bei xgoal-code-Generation\n");
#ifdef space
					gen_soploplop_code(inst, (int) v.head_vars, (MW) pptr,  (MW) &copyspace_required_goal[v.goal_nr]);
#else
					gen_soplop_code(inst, (int) v.head_vars, (MW) pptr);
#endif
				} /* switch */
			} /* if */
		else
		{
#ifdef space 
			gen_loplop_code(inst, (MW) pptr, (MW) &copyspace_required_goal[v.goal_nr]);
#else
			gen_lop_code(inst, (MW) pptr);
#endif
		}
		} /* if flag_v1_xgoal_generated */

		if ((flag_v1_xgoal_generated == FALSE) && (inst == c_xgoal)) 
				gen_nop_code(i);
		}
	if (arity > 0)
		if ((v.goal_nr < v.headgoal_count) || (bipnr >= 0))
			{
			gen_head_struct(arity, gptr);
			}
		else
				{
				gen_goal_struct(arity, gptr);
				}
	}

#ifdef space
	/* variable-adress-backtracking (copyspace_required) */
	copyspace_required_goal[v.goal_nr] = copyspace_required; 
#endif


v.goal_nr++;
TraceCountGoals(v.goal_count);
if (bipnr < 0)
	if (v.goal_nr == v.goal_count)
		gen_nop_code(c_lastcall);
	else
		gen_nop_code(c_call);
else if (functor == ExitBlockFunctor)
	gen_nop_code(c_exit_block);
else if (functor == BlockFunctor)
	{
	if (v.goal_nr++ == v.headgoal_count)
		gen_sopsop_code(c_block_in, (int) v.head_vars, g_meta);
	else
		gen_sop_code(c_block_in, g_meta);
	gen_sop_code(c_block_in, (int) -((BlockSize + 2) * ((int)SIZEMW / (int)SIZECHAR)));
	gen_nop_code(c_call);
	gen_sop_code(c_block_out, (int) -((BlockSize + 2) * ((int)SIZEMW / (int)SIZECHAR)));
	if (v.goal_nr == v.goal_count)
		gen_nop_code(c_lastcall);
	else
		gen_nop_code(c_call);
	}
else if (v.goal_nr == v.goal_count)
	gen_nop_code(c_nogoal);
}


/* generates the intermediate-code for a structure sptr with arity arity */
static void gen_goal_struct(MW arity, MWPTR sptr)
{
MW      m;
MWPTR   mptr, functor;
int     i;

while (--arity >= 0)
	{
	m = *(mptr = sptr++);
	StructDeref(m, mptr);
	if (RefUndef(m))
		{
		i = find_var(mptr);
		if (v.var[i].type == VOID_VAR)
			{
			copyspace_required++; /* copyspace required for void_variable */
			TraceCopySpace_required;
			gen_nop_code(g_void);
			}
		else
			if (v.var[i].var_nr++ == 0)
				{	
				copyspace_required++; /* copyspace required for a first_variable */
				TraceCopySpace_required;

				switch ((int) v.var[i].offset)
					{
					case Offset(0):
						gen_nop_code(g_0fstvar);
						break;
					case Offset(1):
						gen_nop_code(g_1fstvar);
						break;
					case Offset(2):
						gen_nop_code(g_2fstvar);
						break;
					case Offset(3):
						gen_nop_code(g_3fstvar);
						break;
					default:
						v.var[i].tempadress = -1;
						gen_sop_code(g_fstvar, (int) v.var[i].offset);
					}
				} 
			else
				switch ((int) v.var[i].offset)
					{
					case Offset(0):
						gen_nop_code(g_0nxtvar);
						break;
					case Offset(1):
						gen_nop_code(g_1nxtvar);
						break;
					case Offset(2):
						gen_nop_code(g_2nxtvar);
						break;
					case Offset(3):
						gen_nop_code(g_3nxtvar);
						break;
					default:
						if (v.var[i].tempadress == -2)
							gen_sop_code(g_nxtvar, (int) v.var[i].offset);
						else 
						{
#ifdef variables_right_order_writing
							gen_sop_code(g_nxtvar, (int) v.var[i].offset);
#else
							gen_sop_code(g_nxtvar, (int) v.var[i].offset);
#endif
						}
					}
		}
	else if (Nil(m))
		gen_nop_code(g_nil);
	else if(Int(m) || Atom(m))
		gen_lop_code(g_const, m);
	else if (List(m))
		{
		gen_nop_code(g_list);
		copyspace_required += 2; /* copyspace required for a list */
		TraceCopySpace_required;
		gen_goal_struct(2, XtrList(m));
		}
	else /* Struct(m) */
		{
		mptr = XtrStruct(m);
		functor = XtrFunctor(mptr);
		gen_lop_code(g_struct, (MW) functor);
		copyspace_required += FunctArity(functor); /* copyspace required for a structure */
		TraceCopySpace_required;
		gen_goal_struct(FunctArity(functor), FirstArgPtr(mptr));
		}
	} 
}


static void gen_nop_code(int instr)
{
#ifdef zwischencodeausgabe
	fprintf(zfile, "codeptr: %lx instr: %u\n",codeptr,instr);
#endif
if (compile_V1_head) 
	headendptr = codeptr;
else
	goalendptr = codeptr;
InstrStore(codeptr, instr);
CodeStackCheck;
}


static void gen_sop_code(int instr, int arg1)
{
#ifdef zwischencodeausgabe
	fprintf(zfile, "codeptr: %lx instr: %u; arg1: %d \n",codeptr,instr, arg1);
#endif
if (compile_V1_head) 
	headendptr = codeptr;
else
	goalendptr = codeptr;
InstrStore(codeptr, instr);
SopStore(codeptr, arg1);
CodeStackCheck;
}


static void gen_lop_code(int instr, MW arg1)
{
#ifdef zwischencodeausgabe
	fprintf(zfile, "codeptr: %lx instr: %u; arg1: %lx \n",codeptr,instr, arg1);
#endif
if (compile_V1_head) 
	headendptr = codeptr;
else
	goalendptr = codeptr;
InstrStore(codeptr, instr);
LopStore(codeptr, arg1);
CodeStackCheck;
}


static void gen_sopsop_code(int instr, int arg1, int arg2)
{
#ifdef zwischencodeausgabe
	fprintf(zfile, "codeptr: %lx instr: %u; arg1: %d arg2: %d\n",codeptr,instr, arg1, arg2);
#endif
if (compile_V1_head) 
	headendptr = codeptr;
else
	goalendptr = codeptr;
InstrStore(codeptr, instr);
SopStore(codeptr, arg1);
SopStore(codeptr, arg2);
CodeStackCheck;
}


static void gen_soplop_code(int instr, int arg1, MW arg2)
{
#ifdef zwischencodeausgabe
	fprintf(zfile, "codeptr: %lx instr: %u; arg1: %d arg2: %lx\n",codeptr,instr, arg1, arg2);
#endif
if (compile_V1_head) 
	headendptr = codeptr;
else
	goalendptr = codeptr;
InstrStore(codeptr, instr);
SopStore(codeptr, arg1);
LopStore(codeptr, arg2);
CodeStackCheck;
}


static void gen_loplop_code(int instr, MW arg1, MW arg2)
{
#ifdef zwischencodeausgabe
	fprintf(zfile, "codeptr: %lx instr: %u; arg1: %lx arg2: %lx\n",codeptr,instr, arg1, arg2);
#endif
if (compile_V1_head) 
	headendptr = codeptr;
else
	goalendptr = codeptr;
InstrStore(codeptr, instr);
LopStore(codeptr, arg1);
LopStore(codeptr, arg2);
CodeStackCheck;
}

static void gen_soploplop_code(int instr, int arg1, MW arg2, MW arg3)
{
#ifdef zwischencodeausgabe
	fprintf(zfile, "codeptr: %lx instr: %u; arg1: %d arg2: %lx arg3: %lx\n",codeptr,instr, arg1, arg2, arg3);
#endif
if (compile_V1_head) 
	headendptr = codeptr;
else
	goalendptr = codeptr;
InstrStore(codeptr, instr);
SopStore(codeptr, arg1);
LopStore(codeptr, arg2);
LopStore(codeptr, arg3);
CodeStackCheck;
}

static void gen_sopsoploplop_code(int instr, int arg1, MW arg2, MW arg3, MW arg4)
{
#ifdef zwischencodeausgabe
	fprintf(zfile, "codeptr: %lx instr: %u; arg1: %d arg2: %lx arg3: %lx arg4: %lx\n",codeptr,instr, arg1, arg2, arg3, arg4);
#endif
if (compile_V1_head) 
	headendptr = codeptr;
else
	goalendptr = codeptr;
InstrStore(codeptr, instr);
SopStore(codeptr, arg1);
LopStore(codeptr, arg2);
LopStore(codeptr, arg3);
LopStore(codeptr, arg4);
CodeStackCheck;
}

static void gen_loploplopsop_code(int instr, MW arg1, MW arg2, MW arg3, int arg4)
{
#ifdef zwischencodeausgabe
	fprintf(zfile, "codeptr: %lx instr: %u; arg1: %lx arg2: %lx arg3: %lx arg4: %u\n",codeptr,instr, arg1, arg2, arg3, arg4);
#endif
if (compile_V1_head) 
	headendptr = codeptr;
else
	goalendptr = codeptr;
InstrStore(codeptr, instr);
LopStore(codeptr, arg1);
LopStore(codeptr, arg2);
LopStore(codeptr, arg3);
SopStore(codeptr, arg4);
CodeStackCheck;
}

static void gen_soploploplopsop_code(int instr, int arg1, MW arg2, MW arg3, MW arg4, int arg5)
{
#ifdef zwischencodeausgabe
	fprintf(zfile, "codeptr: %lx instr: %u; arg1: %d arg2: %u arg3: %lx arg4: %lx arg5: %u\n",codeptr,instr, arg1, arg2, arg3, arg4,arg5);
#endif
if (compile_V1_head)
	headendptr = codeptr;
else
	goalendptr = codeptr;
InstrStore(codeptr, instr);
SopStore(codeptr, arg1);
LopStore(codeptr, arg2);
LopStore(codeptr, arg3);
LopStore(codeptr, arg4);
SopStore(codeptr, arg5);
CodeStackCheck;
}

/* decompiles a clause at adress clauseptr with the functor functor */
int clause_decompile(MWPTR functor, CODEPTR clauseptr, MWPTR termptr)
{
MW      m;
MWPTR   mptr;
int     err;

ErrorNumber = setjmp(compiler_err);
if (ErrorNumber)
	{
	return(ErrorNumber);
	}
codeptr = clauseptr;
mptr = termptr;
if (FunctArity(functor) == 0)
	*termptr = MakeAtom(FunctIndex(functor));
else
	{
	CreateStruct(termptr, functor, FunctArity(functor));
	if ((err = h_struct_decompile(termptr, FunctArity(functor))) != Success)
		return(err);
	}
if (InstrRead(codeptr) == c_nogoal)
	return(Success);
first_goal = TRUE;
m = *mptr;
CreateStruct(mptr, RuleFunctor, 2);
*mptr++ = m;
while (goal_decompile(mptr))
	{
	m = *mptr;
	CreateStruct(mptr, CommaFunctor, 2);
	*mptr++ = m;
	}
return(Success);
}


/* decompiles a goal of a clause at the adress gptr */
static int goal_decompile(MWPTR gptr)
{
MW      m, arity;
MWPTR   pptr, functor;
int     i, err;

switch (InstrFetch(codeptr))
	{
	case c_1_goal:
	case c_2_goal:
	case c_3_goal:
	case c_4_goal:
		goto goal_decompile_0;
	case c_goal:
		if (first_goal)
			{
			SIncr(codeptr);
goal_decompile_0:
			first_goal = FALSE;
			}
		LopFetch(m, codeptr);
		pptr = (MWPTR) m;
		functor = ProcFunctor(pptr);
		if (FunctArity(functor) == 0)
			*gptr = MakeAtom(FunctIndex(functor));
		else
			{
			CreateStruct(gptr, functor, FunctArity(functor));
			if ((err = g_struct_decompile(gptr, FunctArity(functor))) != Success)
				return(err);
			}
		if (InstrFetch(codeptr) == c_call)
			return(TRUE);
		return(FALSE);
	case c_1_xgoal:
	case c_2_xgoal:
	case c_3_xgoal:
	case c_4_xgoal:
		goto goal_decompile_1;
	case c_xgoal:
		if (first_goal)
			{
			SIncr(codeptr);
goal_decompile_1:
			first_goal = FALSE;
			}
		LopFetch(m, codeptr);
		pptr = (MWPTR) m;
		functor = ProcFunctor(pptr);
		CreateStruct(gptr, functor, FunctArity(functor));
		*gptr++ = Var(codeptr);
		if ((err = g_struct_decompile(gptr, FunctArity(functor) - 1)) != Success)
			return(err);
		if (InstrFetch(codeptr) == c_call)
			return(TRUE);
		return(FALSE);
	case c_cut:
		*gptr = MakeAtom(CutAtomIndex);
		if (! first_goal)
			SIncr(codeptr);
		if (InstrRead(codeptr) != c_nogoal)
			return(TRUE);
		return(FALSE);
	case c_builtin:
		i = SopFetch(codeptr);
		m = builtin[i].bt_name;
		arity = builtin[i].bt_arity;
		if (arity == 0)
			*gptr = MakeAtom(m);
		else
			{
			CreateStruct(gptr, find_func(m, arity), arity);
			if ((err = h_struct_decompile(gptr, arity)) != Success)
				return(err);
			}
		switch (InstrRead(codeptr))
			{
			case c_block_in:
				SIncr(codeptr);             /* c_block_in       */
				if (first_goal)
					{
					first_goal = FALSE;
					SIncr(codeptr);
					}
				SIncr(codeptr);             /* g_meta           */
				SIncr(codeptr);             /* c_block_in       */
				SIncr(codeptr);             /* var_offset       */
				SIncr(codeptr);             /* c_call           */
				SIncr(codeptr);             /* c_block_out      */
				SIncr(codeptr);             /* var_offset       */
				if (InstrFetch(codeptr) == c_call)
					return(TRUE);
				/* fall through to c_exit_block */
			case c_nogoal:
				return(FALSE);
			default:
				return(TRUE);
			}
	case c_metacall:
		if (first_goal)
			{
			first_goal = FALSE;
			SIncr(codeptr);
			}
		SIncr(codeptr);
		SIncr(codeptr);
		*gptr = Var(codeptr);
		if (InstrFetch(codeptr) == c_call)
			return(TRUE);
		return(FALSE);
	case c_exit_block:
		*gptr = MakeAtom(ExitBlockAtomIndex);
		return(FALSE);
	default:
		longjmp(compiler_err, SystemError);
		return(FALSE);
	}
}


/* decompiles a structure of a clause at the adress gptr with arity arity*/
static int h_struct_decompile(MWPTR gptr, MW arity)
{
MW      m;
MWPTR   mptr;
MW      i;
int     err;

while (--arity >= 0)
	switch (HeadMux(InstrFetch(codeptr)))
		{
		case HeadMux(h_const):							/* long */
			LopFetch(m, codeptr);
			*gptr++ = m;
			break;
		case HeadMux(h_nil):							/* no operand */
			*gptr++ = MakeNil;
			break;
		case HeadMux(h_list):							/* list - 2 elements */
			*gptr++ = MakeList(copyptr -= 2);                              /* stackchange: no modification */
			if ((err = h_struct_decompile(copyptr, 2)) != Success)         /* stackchange: no modification */
				return(err);
			break;
		case HeadMux(h_struct):							/* structure - functor and n elements */
			mptr = gptr++;
			FetchStruct(mptr, codeptr, i = ArityFetch(codeptr));
			if ((err = h_struct_decompile(mptr, i)) != Success)
				return(err);
			break;
		case HeadMux(h_fstvar):							/* local variable */
			Var(codeptr) = MakeRef(gptr);
		case HeadMux(h_void):							/* void */
			*gptr++ = MakeUndef(gptr);
			break;
		case HeadMux(h_nxtvar):							/* local variable */
			*gptr++ = Var(codeptr);
			break;
		case HeadMux(h_fsttemp):						/* temporal variable */
			Tmp(codeptr) = MakeRef(gptr);
			*gptr++ = MakeUndef(gptr);
			break;
		case HeadMux(h_nxttemp):						/* temporal variable */
			*gptr++ = Tmp(codeptr);
			break;
		case HeadMux(h_fsttmp0):						/* temporal variable */
			ShortTmp(0) = MakeRef(gptr);
			*gptr++ = MakeUndef(gptr);
			break;
		case HeadMux(h_nxttmp0):						/* temporal variable */
			*gptr++ = ShortTmp(0);
			break;
		default:
			longjmp(compiler_err, SystemError);
		}
return(Success);
}


/* decompiles a structure of a clause at the adress gptr with arity arity*/
static int g_struct_decompile(MWPTR gptr, MW arity)
{
MW      m;
MWPTR   mptr;
MW      i;
int     err;

while (--arity >= 0)
	switch (GoalMux(InstrFetch(codeptr)))
		{
		case GoalMux(g_const):					/* long */
			LopFetch(m, codeptr);
			*gptr++ = m;
			break;
		case GoalMux(g_nil):					/* no operand */
			*gptr++ = MakeNil;
			break;
		case GoalMux(g_list):					/* list - 2 elements */
			*gptr++ = MakeList(copyptr -= 2);                             /* stackchange: no modification */
			if ((err = g_struct_decompile(copyptr, 2)) != Success)        /* stackchange: no modification */
				return(err);
			break;
		case GoalMux(g_struct):					/* structure - functor and n elements */
			mptr = gptr++;
			FetchStruct(mptr, codeptr, i = ArityFetch(codeptr));
			if ((err = g_struct_decompile(mptr, i)) != Success)
				return(err);
			break;
		case GoalMux(g_fstvar):					/* local variable - short*/
			Var(codeptr) = MakeRef(gptr);
		case GoalMux(g_void):					/* void */
			*gptr++ = MakeUndef(gptr);
			break;
		case GoalMux(g_nxtvar):					/* local variable - short*/
			*gptr++ = Var(codeptr);
			break;
		case GoalMux(g_0fstvar):				/* local variable - short*/
			ShortVar(0) = MakeRef(gptr);
			*gptr++ = MakeUndef(gptr);
			break;
		case GoalMux(g_0nxtvar):				/* local variable - short*/
			*gptr++ = ShortVar(0);
			break;
		case GoalMux(g_1fstvar):				/* local variable - short*/
			ShortVar(1) = MakeRef(gptr);
			*gptr++ = MakeUndef(gptr);
			break;
		case GoalMux(g_1nxtvar):				/* local variable - short*/
			*gptr++ = ShortVar(1);
			break;
		case GoalMux(g_2fstvar):				/* local variable - short*/
			ShortVar(2) = MakeRef(gptr);
			*gptr++ = MakeUndef(gptr);
			break;
		case GoalMux(g_2nxtvar):				/* local variable - short*/
			*gptr++ = ShortVar(2);
			break;
		case GoalMux(g_3fstvar):				/* local variable - short*/
			ShortVar(3) = MakeRef(gptr);
			*gptr++ = MakeUndef(gptr);
			break;
		case GoalMux(g_3nxtvar):				/* local variable - short*/
			*gptr++ = ShortVar(3);
			break;
		default:
			longjmp(compiler_err, SystemError);
		}
return(Success);
}


/* inserts the clause arg[0] in the database */
int b_asserta(MW arg[])
{
CODEPTR codeptr;
return(compile_clause(arg[0], TRUE, &codeptr));
}


/* inserts the clause arg[0] in the database (at the end)*/
int b_assertz(MW arg[])
{
CODEPTR codeptr;
return(compile_clause(arg[0], FALSE, &codeptr));
}

/*==================================================================================
	       VAM_1P-Compiler
  ==================================================================================*/
int mi_compiler(CODEPTR headptr, int mode)

{
register CODEPTR    goalptr;			/* current goalptr */
register MWPTR      m1ptr;
register MW         m1;
register MW         i;
register MWPTR      m2ptr;
register MW         m2;
register CHARPTR    headframeptr;		/* current heaframeptr */
register CHARPTR    goalframeptr;		/* current goalframeptr */
register MW			temp0;
		 MWPTR      m3ptr;
		 int        bipnr;				/* number of current builtin */
		 int        nogoal_bip;
		 MW         a[4];
		 MW         ErrorAtomIndex;	
		 MWPTR      ErrorFunctor;
		 MWPTR      berror;
		 MW         berr;
		 MW         m3;
		 MW         temp[64]; 
		 CODE       code[28];
		 CODEPTR    goalptr2;
		 CODEPTR    headptr2;
		 MWPTR      blockptr;
		 MW         ball;
		 CODEPTR    blockoutptr;
		 short      off1, off2;			/* evaluates the offset of an local variable */
		 MWPTR      goalptrtemp, headptrtemp;
		 MW         goaltemp, headtemp;
		 MW         clause_name_head, clause_name_lastsubgoal;		
		 MW         clause_arity_head, clause_arity_lastsubgoal;
		 short		var_offset;


/* write VAM1P-commands for the selfrecursive last-goal */
if (mode == write_commands)
	{
	if ((flag_v1_xgoal_generated == FALSE) || (headbeginptr == goalbeginptr))
		return (ErrorNumber = Success); /* no VAM1P-code-generation */
	headptr = headbeginptr;
	goalptr = goalbeginptr;
	v1_codeptr[current_goal] = codeptr; /* backpatching from V1-Code-address */
	current_goal++;
	}	
/* write the V1-xgoal command for the selfrecursive last-goal */
else
	{	
	if (headbeginptr > headendptr)
		return (ErrorNumber = Success); /* no VAM1P-code-generation */
	headptr = headbeginptr;
	goalptr = headbeginptr;
	goto write_xgoal_label;
	}


/* calculate correct offset -> negative adress in the stack frame*/
if ((varoffset >= 0) && (varoffset <= 3))
	var_offset = Offset(varoffset);
else
	var_offset = varoffset;

/* evaluate the nxtvar-command included in the xgoal-command */
	switch(combine_compiler(InstrFetch(headptr), g_nxtvar))
	{
	case combine_compiler(h_const, g_nxtvar): /* varoffset (short), atom/int (long) */
		LopFetch(m1, headptr);
		gen_soplop_code(v1_nxtvar_const, var_offset, m1); 
		break;
	case combine_compiler(h_nil, g_nxtvar): /* varoffset (short), no operand */
		gen_sop_code(v1_nxtvar_nil, var_offset);
		break;
	case combine_compiler(h_list, g_nxtvar):/* varoffset (short) */
		gen_sop_code(v1_nxtvar_list, var_offset);
		i = 2;
		goto compile_head_unify;
	case combine_compiler(h_struct, g_nxtvar):/* functor (long) and n operands, varoffset(short) */
		m1 = (MW) FunctorFetch(headptr);
		gen_soplop_code(v1_nxtvar_struct, var_offset, m1);
		i = ArityFetch(headptr);
		FirstArgIncr(headptr);
		goto compile_head_unify;
	case combine_compiler(h_void, g_nxtvar):/* no operand, varoffset (short) */
		break;
	case combine_compiler(h_fstvar, g_nxtvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		if (off1 != var_offset)
			gen_sopsop_code(v1_fstvar_nxtvar, off1, var_offset); 
		break;
	case combine_compiler(h_nxtvar, g_nxtvar): /* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		if (off1 != var_offset)
			gen_sopsop_code(v1_nxtvar_nxtvar, off1, var_offset);
		break;
	case combine_compiler(h_fsttemp, g_nxtvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		gen_sopsop_code(v1_fsttmp_nxtvar,  off1, var_offset);
		break;
	case combine_compiler(h_nxttemp, g_nxtvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		gen_sopsop_code(v1_nxtvar_nxttmp, var_offset, off1);
		break;
	case combine_compiler(h_fsttmp0, g_nxtvar):/*  varoffset (short) */
		gen_sop_code(v1_fsttmp0_nxtvar, var_offset); 
		break;
	case combine_compiler(h_nxttmp0, g_nxtvar):/* varoffset (short) */
		gen_sop_code(v1_nxtvar_nxttmp0, var_offset); 
		break;
	case combine_compiler(h_undefproc, g_nxtvar):
		SDecr(goalptr);
		LDecr(goalptr);
		fprintf(stderr, "Fehler bei CompileUndefProc\n");
		return(ErrorNumber = UndefProcError);
	}


write_commands_label:

switch (combine_compiler(InstrFetch(headptr),InstrFetch(goalptr)))
	{
	case combine_compiler(h_const, g_const):  /* atom/int (long), atom/int (long) */
		LopFetch(m1, headptr);
		LopFetch(m2, goalptr);
		if (m1 == m2)
			{
			/* no code generation */
			goto compile;
			}
		goto compile_fail;
	case combine_compiler(h_const, g_nil):    /* atom/int (long), no operand */
		goto compile_fail;
	case combine_compiler(h_const, g_list):   /* atom/int (long), 2 operands */
		goto compile_fail;
	case combine_compiler(h_const, g_struct): /* atom/int (long), functor (long) and n operands */
		goto compile_fail;
	case combine_compiler(h_const, g_void):   /* atom/int (long), no operand */
		LIncr(headptr);
		goto compile;
	case combine_compiler(h_const, g_fstvar): /* varoffset (short), atom/int (long) */
		LopFetch(m1, headptr);
		gen_soplop_code(v1_fstvar_const, SopFetch(goalptr), m1); 
		goto compile;
	case combine_compiler(h_const, g_nxtvar): /* varoffset (short), atom/int (long) */
		LopFetch(m1, headptr);
		gen_soplop_code(v1_nxtvar_const, SopFetch(goalptr), m1); 
		goto compile;
	case combine_compiler(h_const, g_0fstvar): /* varoffset (short), atom/int (long) */
		LopFetch(m1, headptr);
		gen_soplop_code(v1_fstvar_const, (SHORT) Offset(0), m1); 
		goto compile;
	case combine_compiler(h_const, g_0nxtvar): /* varoffset (short), atom/int (long) */
		LopFetch(m1, headptr);
		gen_soplop_code(v1_nxtvar_const, (SHORT) Offset(0), m1); 
		goto compile;
	case combine_compiler(h_const, g_1fstvar): /* varoffset (short), atom/int (long) */
		LopFetch(m1, headptr);
		gen_soplop_code(v1_fstvar_const, (SHORT) Offset(1), m1); 
		goto compile;
	case combine_compiler(h_const, g_1nxtvar): /* varoffset (short), atom/int (long) */
		LopFetch(m1, headptr);
		gen_soplop_code(v1_nxtvar_const, (SHORT) Offset(1), m1); 
		goto compile;
	case combine_compiler(h_const, g_2fstvar): /* varoffset (short), atom/int (long) */
		LopFetch(m1, headptr);
		gen_soplop_code(v1_fstvar_const, (SHORT) Offset(2), m1); 
		goto compile;
	case combine_compiler(h_const, g_2nxtvar): /* varoffset (short), atom/int (long) */
		LopFetch(m1, headptr);
		gen_soplop_code(v1_nxtvar_const, (SHORT) Offset(2), m1); 
		goto compile;
	case combine_compiler(h_const, g_3fstvar): /* varoffset (short), atom/int (long) */
		LopFetch(m1, headptr);
		gen_soplop_code(v1_fstvar_const, (SHORT) Offset(3), m1); 
		goto compile;
	case combine_compiler(h_const, g_3nxtvar): /* varoffset (short), atom/int (long) */
		LopFetch(m1, headptr);
		gen_soplop_code(v1_nxtvar_const, (SHORT) Offset(3), m1); 
		goto compile;
	case combine_compiler(h_nil, g_const): /* no operand, atom/int (long) */
		goto compile_fail;
	case combine_compiler(h_nil, g_nil): /* no operand, no operand */
		goto compile;
	case combine_compiler(h_nil, g_list): /* no operand, 2 operands */
		goto compile_fail;
	case combine_compiler(h_nil, g_struct): /* no operand, functor (long) n operands */
		goto compile_fail;
	case combine_compiler(h_nil, g_void): /* no operand, no operand */
		goto compile;
	case combine_compiler(h_nil, g_fstvar): /* no operand, varoffset (short) */
		gen_sop_code(v1_fstvar_nil, SopFetch(goalptr));
		goto compile;

	case combine_compiler(h_nil, g_nxtvar): /* varoffset (short), no operand */
		gen_sop_code(v1_nxtvar_nil, SopFetch(goalptr));
		goto compile;
	case combine_compiler(h_nil, g_0fstvar):/* varoffset (short), no operand */
		gen_sop_code(v1_fstvar_nil,(SHORT) Offset(0));
		goto compile;
	case combine_compiler(h_nil, g_0nxtvar):/* varoffset (short), no operand */
		gen_sop_code(v1_nxtvar_nil,(SHORT) Offset(0));
		goto compile;
	case combine_compiler(h_nil, g_1fstvar):/* varoffset (short), no operand */
		gen_sop_code(v1_fstvar_nil,(SHORT) Offset(1));
		goto compile;
	case combine_compiler(h_nil, g_1nxtvar):/* varoffset (short), no operand */
		gen_sop_code(v1_nxtvar_nil,(SHORT) Offset(1));
		goto compile;
	case combine_compiler(h_nil, g_2fstvar):/* varoffset (short), no operand */
		gen_sop_code(v1_fstvar_nil,(SHORT) Offset(2));
		goto compile;
	case combine_compiler(h_nil, g_2nxtvar):/* varoffset (short), no operand */
		gen_sop_code(v1_nxtvar_nil,(SHORT) Offset(2));
		goto compile;
	case combine_compiler(h_nil, g_3fstvar):/* varoffset (short), no operand */
		gen_sop_code(v1_fstvar_nil,(SHORT) Offset(3));
		goto compile;
	case combine_compiler(h_nil, g_3nxtvar):/* varoffset (short), no operand */
		gen_sop_code(v1_nxtvar_nil,(SHORT) Offset(3));
		goto compile;

	case combine_compiler(h_list, g_const):/* 2 operands, atom/int (long) */
		goto compile_fail;
	case combine_compiler(h_list, g_nil):/* 2 operands, no operand */
		goto compile_fail;
	case combine_compiler(h_list, g_list):/* 2 operands, 2 operands */
		goto compile;
	case combine_compiler(h_list, g_struct):/* 2 operands, functor (long) and n operands */
		goto compile_fail;
	case combine_compiler(h_list, g_void): /* 2 operands, no operand */
		i = 2;
compile_head_list_skip: 
		while (--i >= 0)
			switch(HeadMux(InstrFetch(headptr)))
				{
				case HeadMux(h_const):     /* long */
					LIncr(headptr);
				case HeadMux(h_nil):
					break;
				case HeadMux(h_list):		/* list - 2 elements */
					i += 2;
					break;
				case HeadMux(h_struct):		/* structure - functor and n elements */
					i += ArityFetch(headptr);
					FirstArgIncr(headptr);
				case HeadMux(h_void):		/* void - no operand */
					break;
				case HeadMux(h_fstvar):		/* short */
					gen_sop_code(v1_fstvar_void,SopFetch(headptr));
					break;
				case HeadMux(h_nxtvar):		/* short */
				case HeadMux(h_nxttemp):	/* short */
					SIncr(headptr);
				case HeadMux(h_nxttmp0):	/* short */
					break;
				case HeadMux(h_fsttemp):	/* short */
					gen_sop_code(v1_fsttmp_void,SopFetch(headptr));
					break;
				case HeadMux(h_fsttmp0):	/* short */
					gen_nop_code(v1_fsttmp0_void); 
					break;
				default:
					fprintf(stderr, "CompileHeadListSkipError\n");
					return(ErrorNumber = SystemError);
				}
	goto compile; 

	case combine_compiler(h_list, g_fstvar): /*  varoffset (short) */
		gen_sop_code(v1_fstvar_list, SopFetch(goalptr));
		i = 2;
		goto compile_head_unify;
	case combine_compiler(h_list, g_nxtvar):/* varoffset (short) */
		gen_sop_code(v1_nxtvar_list, SopFetch(goalptr));
		i = 2;
		goto compile_head_unify;
	case combine_compiler(h_list, g_0fstvar):/* varoffset (short) */
		gen_sop_code(v1_fstvar_list, (SHORT) Offset(0));
		i = 2;
		goto compile_head_unify;
	case combine_compiler(h_list, g_0nxtvar):/* varoffset (short) */
		gen_sop_code(v1_nxtvar_list, (SHORT) Offset(0));
		i = 2;
		goto compile_head_unify;
	case combine_compiler(h_list, g_1fstvar):/* varoffset (short) */
		gen_sop_code(v1_fstvar_list, (SHORT) Offset(1));
		i = 2;
		goto compile_head_unify;
	case combine_compiler(h_list, g_1nxtvar):/* varoffset (short) */
		gen_sop_code(v1_nxtvar_list, (SHORT) Offset(1));
		i = 2;
		goto compile_head_unify;
	case combine_compiler(h_list, g_2fstvar):/* varoffset (short) */
		gen_sop_code(v1_fstvar_list, (SHORT) Offset(2));
		i = 2;
		goto compile_head_unify;
	case combine_compiler(h_list, g_2nxtvar):/* varoffset (short) */
		gen_sop_code(v1_nxtvar_list, (SHORT) Offset(2));
		i = 2;
		goto compile_head_unify;
	case combine_compiler(h_list, g_3fstvar):/* varoffset (short) */
		gen_sop_code(v1_fstvar_list, (SHORT) Offset(3));
		i = 2;
		goto compile_head_unify;
	case combine_compiler(h_list, g_3nxtvar):/* varoffset (short) */
		gen_sop_code(v1_nxtvar_list, (SHORT) Offset(3));
		i = 2;
		goto compile_head_unify;
	case combine_compiler(h_struct, g_const):/* functor (long) and n operands, atom/int (long) */
		goto compile_fail;
	case combine_compiler(h_struct, g_nil):/* functor (long) and n operands, no operand */
		goto compile_fail;
	case combine_compiler(h_struct, g_list):/* functor (long) and n operands, 2 operands */
		goto compile_fail;
	case combine_compiler(h_struct, g_struct): /* functor (long) and n operands, functor (long) and n operands*/
		CompileCodeStructUnify(headptr, goalptr);
		goto compile;
	case combine_compiler(h_struct, g_void): /* functor (long) and n operands, no operand */
		i = ArityFetch(headptr);
		FirstArgIncr(headptr);
		goto compile_head_list_skip;
	case combine_compiler(h_struct, g_fstvar):/* varoffset(short), functor (long) and n operands  */
		m1 = (MW) FunctorFetch(headptr);
		gen_soplop_code(v1_fstvar_struct, SopFetch(goalptr), m1);
		i = ArityFetch(headptr);
		FirstArgIncr(headptr);
		goto compile_head_unify;
	case combine_compiler(h_struct, g_nxtvar):/* functor (long) and n operands, varoffset(short) */
		m1 = (MW) FunctorFetch(headptr);
		gen_soplop_code(v1_nxtvar_struct, SopFetch(goalptr), m1);
		i = ArityFetch(headptr);
		FirstArgIncr(headptr);
		goto compile_head_unify;
	case combine_compiler(h_struct, g_0fstvar):/* functor (long) and n operands, no operand */
		m1 = (MW) FunctorFetch(headptr);
		gen_soplop_code(v1_fstvar_struct, (SHORT) Offset(0), m1);
		i = ArityFetch(headptr);
		FirstArgIncr(headptr);
		goto compile_head_unify;
	case combine_compiler(h_struct, g_0nxtvar):/* functor (long) and n operands, no operand */
		m1 = (MW) FunctorFetch(headptr);
		gen_soplop_code(v1_nxtvar_struct, (SHORT) Offset(0), m1);
		i = ArityFetch(headptr);
		FirstArgIncr(headptr);
		goto compile_head_unify;
	case combine_compiler(h_struct, g_1fstvar):/* functor (long) and n operands, no operand */
		m1 = (MW) FunctorFetch(headptr);
		gen_soplop_code(v1_fstvar_struct, (SHORT) Offset(1), m1);
		i = ArityFetch(headptr);
		FirstArgIncr(headptr);
		goto compile_head_unify;
	case combine_compiler(h_struct, g_1nxtvar):/* functor (long) and n operands, no operand */
		m1 = (MW) FunctorFetch(headptr);
		gen_soplop_code(v1_nxtvar_struct, (SHORT) Offset(1), m1);
		i = ArityFetch(headptr);
		FirstArgIncr(headptr);
		goto compile_head_unify;
	case combine_compiler(h_struct, g_2fstvar):/* functor (long) and n operands, no operand */
		m1 = (MW) FunctorFetch(headptr);
		gen_soplop_code(v1_fstvar_struct, (SHORT) Offset(2), m1);
		i = ArityFetch(headptr);
		FirstArgIncr(headptr);
		goto compile_head_unify;
	case combine_compiler(h_struct, g_2nxtvar):/* functor (long) and n operands, no operand */
		m1 = (MW) FunctorFetch(headptr);
		gen_soplop_code(v1_nxtvar_struct, (SHORT) Offset(2), m1);
		i = ArityFetch(headptr);
		FirstArgIncr(headptr);
		goto compile_head_unify;
	case combine_compiler(h_struct, g_3fstvar):/* functor (long) and n operands, no operand */
		m1 = (MW) FunctorFetch(headptr);
		gen_soplop_code(v1_fstvar_struct, (SHORT) Offset(3), m1);
		i = ArityFetch(headptr);
		FirstArgIncr(headptr);
		goto compile_head_unify;
	case combine_compiler(h_struct, g_3nxtvar):/* functor (long) and n operands, no operand */
		m1 = (MW) FunctorFetch(headptr);
		gen_soplop_code(v1_nxtvar_struct, (SHORT) Offset(3), m1);
		i = ArityFetch(headptr);
		FirstArgIncr(headptr);
		goto compile_head_unify;
	case combine_compiler(h_void, g_const):   /* no operand, atom/int (long) */
		LIncr(goalptr);
		goto compile; 
	case combine_compiler(h_void, g_nil):/* no operand, no operand */
		goto compile;
	case combine_compiler(h_void, g_list): /* no operand, 2 operands */
		i = 2;
compile_goal_list_skip:
		while (--i >= 0)
			switch(GoalMux(InstrFetch(goalptr)))
				{
				case GoalMux(g_const):		/* long */
					LIncr(goalptr);
				case GoalMux(g_nil):		/* no operand */
					break;
				case GoalMux(g_list):		/* list - 2 elements */
					i += 2;
					break;
				case GoalMux(g_struct):			/* structure - functor and n elements */
					i += ArityFetch(goalptr);
					FirstArgIncr(goalptr);
				case GoalMux(g_void):			/* no operand */
					break;
				case GoalMux(g_fstvar):				/* short */
					gen_sop_code(v1_fstvar_void, SopFetch(goalptr)); 
					break;
				case GoalMux(g_nxtvar):				/* short */
					SIncr(goalptr);
					break;
				case GoalMux(g_0fstvar):				/* short */	
					gen_sop_code(v1_fstvar_void, (SHORT) Offset(0)); 
					break;
				case GoalMux(g_1fstvar):				/* short */	
					gen_sop_code(v1_fstvar_void, (SHORT) Offset(1)); 
					break;
				case GoalMux(g_2fstvar):				/* short */	
					gen_sop_code(v1_fstvar_void, (SHORT) Offset(2)); 
					break;
				case GoalMux(g_3fstvar):				/* short */	
					gen_sop_code(v1_fstvar_void, (SHORT) Offset(3)); 
					break;
				case GoalMux(g_0nxtvar):
				case GoalMux(g_1nxtvar):
				case GoalMux(g_2nxtvar):
				case GoalMux(g_3nxtvar):
					break;
				default:
					fprintf(stderr, "CompileGoalListSkipError\n");
					return(ErrorNumber = SystemError);
				}
			goto compile;
	case combine_compiler(h_void, g_struct):/* no operand, functor (long) n operands */
		i = ArityFetch(goalptr);
		FirstArgIncr(goalptr);
		goto compile_goal_list_skip;
	case combine_compiler(h_void, g_void):/* no operand, no operand */
		goto compile;
	case combine_compiler(h_void, g_fstvar):/* varoffset (short), no operand  */
		gen_sop_code(v1_fstvar_void, SopFetch(goalptr));
		goto compile;
	case combine_compiler(h_void, g_nxtvar):/* no operand, varoffset (short) */
		SIncr(goalptr);
		goto compile; /* no code generation */
	case combine_compiler(h_void, g_0fstvar):/* no operand, no operand */
		gen_sop_code(v1_fstvar_void, (SHORT) Offset(0)); /* code generation */
		goto compile;
	case combine_compiler(h_void, g_1fstvar):/* varoffset (short), no operand */
		gen_sop_code(v1_fstvar_void, (SHORT) Offset(1)); /* code generation */
		goto compile;
	case combine_compiler(h_void, g_2fstvar):/* varoffset (short), no operand */
		gen_sop_code(v1_fstvar_void, (SHORT) Offset(2)); /* code generation */
		goto compile;
	case combine_compiler(h_void, g_3fstvar):/* varoffset (short), no operand */
		gen_sop_code(v1_fstvar_void, (SHORT) Offset(3)); /* code generation */
		goto compile;
	case combine_compiler(h_void, g_0nxtvar):/* no operand, no operand */
	case combine_compiler(h_void, g_1nxtvar):/* no operand, no operand */
	case combine_compiler(h_void, g_2nxtvar):/* no operand, no operand */
	case combine_compiler(h_void, g_3nxtvar):/* no operand, no operand */
		goto compile;                    /* no code generation */
	case combine_compiler(h_fstvar, g_const):  /* varoffset (short), atom/int (long) */
		LopFetch(m1, goalptr);
		gen_soplop_code(v1_fstvar_const, SopFetch(headptr), m1);
		goto compile;
	case combine_compiler(h_fstvar, g_nil): /* varoffset (short), no operand */
		gen_sop_code(v1_fstvar_nil, SopFetch(headptr));
		goto compile;
	case combine_compiler(h_fstvar, g_list): /* varoffset (short), 2 operands */
		gen_sop_code(v1_fstvar_list, SopFetch(headptr));
		i = 2;
		goto compile_goal_unify;
	case combine_compiler(h_fstvar, g_struct): /* varoffset (short), functor (long) and n operands */
		m1 = (MW) FunctorFetch(goalptr);
		gen_soplop_code(v1_fstvar_struct, SopFetch(headptr), m1);
		i = ArityFetch(goalptr);
		FirstArgIncr(goalptr);
		goto compile_goal_unify;
	case combine_compiler(h_fstvar, g_void):/* varoffset (short), no operand */
		gen_sop_code(v1_fstvar_void, SopFetch(headptr));
		goto compile;
	case combine_compiler(h_fstvar, g_fstvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		off2 = SopFetch(goalptr);
		if (off1 != off2)
			gen_sopsop_code(v1_fstvar_fstvar, off1, off2);
		goto compile;
	case combine_compiler(h_fstvar, g_nxtvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		off2 = SopFetch(goalptr);
		if (off1 != off2)
			gen_sopsop_code(v1_fstvar_nxtvar, off1, off2); 
		goto compile;
	case combine_compiler(h_fstvar, g_0fstvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		if (off1 != Offset(0))
			gen_sopsop_code(v1_fstvar_fstvar, off1, (SHORT) Offset(0));
		goto compile;
	case combine_compiler(h_fstvar, g_0nxtvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		if (off1 != Offset(0)) 
			gen_sopsop_code(v1_fstvar_nxtvar, off1, (SHORT) Offset(0)); 
		goto compile;
	case combine_compiler(h_fstvar, g_1fstvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		if (off1 != Offset(1))
			gen_sopsop_code(v1_fstvar_fstvar, off1, (SHORT) Offset(1));
		goto compile;
	case combine_compiler(h_fstvar, g_1nxtvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		if (off1 != Offset(1)) 
			gen_sopsop_code(v1_fstvar_nxtvar, off1, (SHORT) Offset(1)); 
		goto compile;
	case combine_compiler(h_fstvar, g_2fstvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		if (off1 != Offset(2))
			gen_sopsop_code(v1_fstvar_fstvar, off1, (SHORT) Offset(2));
		goto compile;
	case combine_compiler(h_fstvar, g_2nxtvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		if (off1 != Offset(2)) 
			gen_sopsop_code(v1_fstvar_nxtvar, off1, (SHORT) Offset(2)); 
		goto compile;
	case combine_compiler(h_fstvar, g_3fstvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		if (off1 != Offset(3))
			gen_sopsop_code(v1_fstvar_fstvar, off1, (SHORT) Offset(3));
		goto compile;
	case combine_compiler(h_fstvar, g_3nxtvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		if (off1 != Offset(3))
			gen_sopsop_code(v1_fstvar_nxtvar, off1, (SHORT) Offset(3)); 
		goto compile;
	case combine_compiler(h_nxtvar, g_const):/* varoffset (short), varoffset (short) */
		LopFetch(m1, headptr);
		gen_soplop_code(v1_nxtvar_const, SopFetch(headptr), m1); 
		goto compile;
	case combine_compiler(h_nxtvar, g_nil): /* varoffset (short), no operand */
		gen_sop_code(v1_nxtvar_nil, SopFetch(headptr));
		goto compile;
	case combine_compiler(h_nxtvar, g_list):/* varoffset (short), list->2 operands */
		gen_sop_code(v1_nxtvar_list, SopFetch(headptr));
		i = 2;
		goto compile_goal_unify;
	case combine_compiler(h_nxtvar, g_struct):/* varoffset (short), no operand */
		m1 = (MW) FunctorFetch(goalptr);
		gen_soplop_code(v1_nxtvar_struct, SopFetch(headptr), m1);
		i = ArityFetch(goalptr);
		FirstArgIncr(goalptr);
		goto compile_goal_unify;
	case combine_compiler(h_nxtvar, g_void):/* varoffset (short), no operand */
		SIncr(headptr);
		goto compile;
	case combine_compiler(h_nxtvar, g_fstvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		off2 = SopFetch(goalptr);
		if (off1 != off2)
			gen_sopsop_code(v1_fstvar_nxtvar, off1, off2); 
		goto compile;
	case combine_compiler(h_nxtvar, g_nxtvar): /* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		off2 = SopFetch(goalptr);
		if (off1 != off2)
			gen_sopsop_code(v1_nxtvar_nxtvar, off1, off2);
		goto compile;
	case combine_compiler(h_nxtvar, g_0fstvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		if (off1 != Offset(0))
			gen_sopsop_code(v1_fstvar_nxtvar, (SHORT) Offset(0), off1 );
		goto compile;
	case combine_compiler(h_nxtvar, g_0nxtvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		if (off1 != Offset(0))
			gen_sopsop_code(v1_nxtvar_nxtvar, off1, (SHORT) Offset(0));
		goto compile;
	case combine_compiler(h_nxtvar, g_1fstvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		if (off1 != Offset(1))
			gen_sopsop_code(v1_fstvar_nxtvar, (SHORT) Offset(1), off1);
		goto compile;
	case combine_compiler(h_nxtvar, g_1nxtvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		if (off1 != Offset(1))
			gen_sopsop_code(v1_nxtvar_nxtvar, off1, (SHORT) Offset(1));
		goto compile;
	case combine_compiler(h_nxtvar, g_2fstvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		if (off1 != Offset(2))
			gen_sopsop_code(v1_fstvar_nxtvar, (SHORT) Offset(2), off1);
		goto compile;
	case combine_compiler(h_nxtvar, g_2nxtvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		if (off1 != Offset(2))
			gen_sopsop_code(v1_nxtvar_nxtvar, off1, (SHORT) Offset(2));
		goto compile;
	case combine_compiler(h_nxtvar, g_3fstvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		if (off1 != Offset(3))
			gen_sopsop_code(v1_fstvar_nxtvar, (SHORT) Offset(3), off1); 
		goto compile;
	case combine_compiler(h_nxtvar, g_3nxtvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		if (off1 != Offset(3))
			gen_sopsop_code(v1_nxtvar_nxtvar, off1, (SHORT) Offset(3));
		goto compile;
	case combine_compiler(h_fsttemp, g_const):/* varoffset (short), atom/int (long) */
		LopFetch(m1,goalptr);
		gen_sopsop_code(v1_fsttmp_const, SopFetch(headptr), m1);
		goto compile;
	case combine_compiler(h_fsttemp, g_nil):/* varoffset (short), no operand */
		gen_sop_code(v1_fsttmp_nil, SopFetch(headptr));
		goto compile;
	case combine_compiler(h_fsttemp, g_list):/* varoffset (short), list 2 operands */
		gen_sop_code(v1_fsttmp_list, SopFetch(headptr));
		i = 2; 
		goto compile_goal_unify;
	case combine_compiler(h_fsttemp, g_struct):/* varoffset (short), functor long and n operands */
		m1 = (MW) FunctorFetch(goalptr);
		gen_soplop_code(v1_fsttmp_struct, SopFetch(headptr), m1);
		i = ArityFetch(goalptr);
		FirstArgIncr(goalptr);
		goto compile_goal_unify;
	case combine_compiler(h_fsttemp, g_void):/* varoffset (short), no operand */
		gen_sop_code(v1_fsttmp_void, SopFetch(headptr));
		goto compile;
	case combine_compiler(h_fsttemp, g_fstvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		off2 = SopFetch(goalptr);
		gen_sopsop_code(v1_fstvar_fsttmp, off2, off1);
		goto compile;
	case combine_compiler(h_fsttemp, g_nxtvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		off2 = SopFetch(goalptr);
		gen_sopsop_code(v1_fsttmp_nxtvar, off2, off1);
		goto compile;
	case combine_compiler(h_fsttemp, g_0fstvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		gen_sopsop_code(v1_fstvar_fsttmp, (SHORT) Offset(0), off1);
		goto compile;
	case combine_compiler(h_fsttemp, g_0nxtvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		gen_sopsop_code(v1_fsttmp_nxtvar, off1, (SHORT) Offset(0));
		goto compile;
	case combine_compiler(h_fsttemp, g_1fstvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		gen_sopsop_code(v1_fstvar_fsttmp, (SHORT) Offset(1), off1);
		goto compile;
	case combine_compiler(h_fsttemp, g_1nxtvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		gen_sopsop_code(v1_fsttmp_nxtvar, off1, (SHORT) Offset(1));
		goto compile;
	case combine_compiler(h_fsttemp, g_2fstvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		gen_sopsop_code(v1_fstvar_fsttmp, (SHORT) Offset(2), off1);
		goto compile;
	case combine_compiler(h_fsttemp, g_2nxtvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		gen_sopsop_code(v1_fsttmp_nxtvar, off1, (SHORT) Offset(2));
		goto compile;
	case combine_compiler(h_fsttemp, g_3fstvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		gen_sopsop_code(v1_fstvar_fsttmp, (SHORT) Offset(3), off1);
		goto compile;
	case combine_compiler(h_fsttemp, g_3nxtvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		gen_sopsop_code(v1_fsttmp_nxtvar, off1, (SHORT) Offset(3));
		goto compile;
		case combine_compiler(h_nxttemp, g_const):/* varoffset (short), atom/int (long) */
		LopFetch(m1,goalptr);
		gen_soplop_code(v1_nxttmp_const,SopFetch(headptr), m1);
		goto compile;
	case combine_compiler(h_nxttemp, g_nil):/* varoffset (short), no operand */
		gen_sop_code(v1_nxttmp_nil,SopFetch(headptr));
		goto compile;
	case combine_compiler(h_nxttemp, g_list):/* varoffset (short), list - 2 operands */
		gen_sop_code(v1_nxttmp_list,SopFetch(headptr));
		i = 2;
		goto compile_goal_unify;
	case combine_compiler(h_nxttemp, g_struct):/* varoffset (short), functor (long) and n operands */
		m1 = (MW) FunctorFetch(goalptr);
		gen_soplop_code(v1_nxttmp_struct, SopFetch(headptr), m1);
		i = ArityFetch(goalptr);
		FirstArgIncr(goalptr);
		goto compile_goal_unify;
	case combine_compiler(h_nxttemp, g_void):/* varoffset (short), no operand */
		SIncr(headptr);
		goto compile;

	case combine_compiler(h_nxttemp, g_fstvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		off2 = SopFetch(goalptr);
		gen_sopsop_code(v1_fstvar_nxttmp, off2, off1);
		goto compile;
	case combine_compiler(h_nxttemp, g_nxtvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		off2 = SopFetch(goalptr);
		gen_sopsop_code(v1_nxtvar_nxttmp, off2, off1);
		goto compile;
	case combine_compiler(h_nxttemp, g_0fstvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		gen_sopsop_code(v1_fstvar_nxttmp, (SHORT) Offset(0), off1);
		goto compile;
	case combine_compiler(h_nxttemp, g_0nxtvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		gen_sopsop_code(v1_nxtvar_nxttmp, (SHORT) Offset(0), off1);
		goto compile;
	case combine_compiler(h_nxttemp, g_1fstvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		gen_sopsop_code(v1_fstvar_nxttmp, (SHORT) Offset(1), off1);
		goto compile;
	case combine_compiler(h_nxttemp, g_1nxtvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		gen_sopsop_code(v1_nxtvar_nxttmp, (SHORT) Offset(1), off1);
		goto compile;
	case combine_compiler(h_nxttemp, g_2fstvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		gen_sopsop_code(v1_fstvar_nxttmp, (SHORT) Offset(2), off1);
		goto compile;
	case combine_compiler(h_nxttemp, g_2nxtvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		gen_sopsop_code(v1_nxtvar_nxttmp, (SHORT) Offset(2), off1);
		goto compile;
	case combine_compiler(h_nxttemp, g_3fstvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		gen_sopsop_code(v1_fstvar_nxttmp, (SHORT) Offset(3), off1);
		goto compile;
	case combine_compiler(h_nxttemp, g_3nxtvar):/* varoffset (short), varoffset (short) */
		off1 = SopFetch(headptr);
		gen_sopsop_code(v1_nxtvar_nxttmp, (SHORT) Offset(3), off1);
		goto compile;
	case combine_compiler(h_fsttmp0, g_const):/* atom/int long */
		LopFetch(m1, goalptr);
		gen_lop_code(v1_fsttmp0_const, m1); 
		goto compile;
	case combine_compiler(h_fsttmp0, g_nil):/* no operand */
		gen_nop_code(v1_fsttmp0_nil); 
		goto compile;
	case combine_compiler(h_fsttmp0, g_list):/* list- 2 operands */
		gen_nop_code(v1_fsttmp0_list); 
		i = 2;
		goto compile_goal_unify;
	case combine_compiler(h_fsttmp0, g_struct):/* functor (long) and n operands */
		m1 = (MW) FunctorFetch(goalptr);
		gen_lop_code(v1_fsttmp0_struct, m1);
		i = ArityFetch(goalptr);
		FirstArgIncr(goalptr);
		goto compile_goal_unify;
	case combine_compiler(h_fsttmp0, g_void):/* no operand */
		gen_nop_code(v1_fsttmp0_void); 
		goto compile;
	case combine_compiler(h_fsttmp0, g_fstvar):/* varoffset (short) */
		off2 = SopFetch(goalptr);
		gen_sop_code(v1_fstvar_fsttmp0, off2);
		goto compile;
	case combine_compiler(h_fsttmp0, g_nxtvar):/* varoffset (short) */
		off2 = SopFetch(goalptr);
		gen_sop_code(v1_fsttmp0_nxtvar, off2); 
		goto compile;
	case combine_compiler(h_fsttmp0, g_0fstvar):/* varoffset (short) */
		gen_sop_code(v1_fstvar_fsttmp0, (SHORT) Offset(0)); 
		goto compile;
	case combine_compiler(h_fsttmp0, g_0nxtvar):/*  varoffset (short) */
		gen_sop_code(v1_fsttmp0_nxtvar, (SHORT) Offset(0));
		goto compile;
	case combine_compiler(h_fsttmp0, g_1fstvar):/* varoffset (short) */
		gen_sop_code(v1_fstvar_fsttmp0, (SHORT) Offset(1));
		goto compile;
	case combine_compiler(h_fsttmp0, g_1nxtvar):/* varoffset (short) */
		gen_sop_code(v1_fsttmp0_nxtvar, (SHORT) Offset(0));
		goto compile;
	case combine_compiler(h_fsttmp0, g_2fstvar):/* varoffset (short) */
		gen_sop_code(v1_fstvar_fsttmp0, (SHORT) Offset(2));
		goto compile;
	case combine_compiler(h_fsttmp0, g_2nxtvar):/* varoffset (short) */
		gen_sop_code(v1_fsttmp0_nxtvar, (SHORT) Offset(2));
		goto compile;
	case combine_compiler(h_fsttmp0, g_3fstvar):/* varoffset (short) */
		gen_sop_code(v1_fstvar_fsttmp0, (SHORT) Offset(3));
		goto compile;
	case combine_compiler(h_fsttmp0, g_3nxtvar):/* varoffset (short) */
		gen_sop_code(v1_fsttmp0_nxtvar, (SHORT) Offset(3));
		goto compile;
	case combine_compiler(h_nxttmp0, g_const):/* atom/int (long) */
		LopFetch(m1, goalptr);
		gen_lop_code(v1_nxttmp0_const, m1);
		goto compile;
	case combine_compiler(h_nxttmp0, g_nil):/* no operand */
		gen_nop_code(v1_nxttmp0_nil); 
		goto compile;
	case combine_compiler(h_nxttmp0, g_list):/* list-2 operands */
		gen_nop_code(v1_nxttmp0_list); 
		i = 2;
		goto compile_goal_unify;
	case combine_compiler(h_nxttmp0, g_struct):/* functor (long) and n operands */
		m1 = (MW) FunctorFetch(goalptr);
		/* gen_soplop_code(v1_nxttmp0_struct, SopFetch(headptr), m1); */
		gen_lop_code(v1_nxttmp0_struct, m1); 
		i = ArityFetch(goalptr);
		FirstArgIncr(goalptr);
		goto compile_goal_unify;
	case combine_compiler(h_nxttmp0, g_void):/* no operand */
		goto compile;
	case combine_compiler(h_nxttmp0, g_fstvar):/* varoffset (short) */
		off2 = SopFetch(goalptr);
		/* gen_sopsop_code(v1_fstvar_nxttmp0, off2, off1); */
		gen_sop_code(v1_fstvar_nxttmp0, off2);
		goto compile;
	case combine_compiler(h_nxttmp0, g_nxtvar):/* varoffset (short) */
		off2 = SopFetch(goalptr);
		gen_sop_code(v1_nxtvar_nxttmp0, off2);
		goto compile;
	case combine_compiler(h_nxttmp0, g_0fstvar):/* varoffset (short) */
		gen_sop_code(v1_fstvar_nxttmp0, (SHORT) Offset(0));
		goto compile;
	case combine_compiler(h_nxttmp0, g_0nxtvar):/* varoffset (short) */
		gen_sop_code(v1_nxtvar_nxttmp0, (SHORT) Offset(0)); 
		goto compile;
	case combine_compiler(h_nxttmp0, g_1fstvar):/* varoffset (short) */
		gen_sop_code(v1_fstvar_nxttmp0, (SHORT) Offset(1)); 
		goto compile;
	case combine_compiler(h_nxttmp0, g_1nxtvar):/*  varoffset (short) */
		gen_sop_code(v1_nxtvar_nxttmp0, (SHORT) Offset(1)); 
		goto compile;
	case combine_compiler(h_nxttmp0, g_2fstvar):/* varoffset (short) */
		gen_sop_code(v1_fstvar_nxttmp0, (SHORT) Offset(2));
		goto compile;
	case combine_compiler(h_nxttmp0, g_2nxtvar):/* varoffset (short) */
		gen_sop_code(v1_nxtvar_nxttmp0, (SHORT) Offset(2));
		goto compile;
	case combine_compiler(h_nxttmp0, g_3fstvar):/*  varoffset (short) */
		gen_sop_code(v1_fstvar_nxttmp0, (SHORT) Offset(3));
		goto compile;
	case combine_compiler(h_nxttmp0, g_3nxtvar):/* varoffset (short) */
		gen_sop_code(v1_nxtvar_nxttmp0, (SHORT) Offset(3));
		goto compile;


	case combine_compiler(h_undefproc, g_const):
	case combine_compiler(h_undefproc, g_nil):
	case combine_compiler(h_undefproc, g_list):
	case combine_compiler(h_undefproc, g_struct):
	case combine_compiler(h_undefproc, g_void):
	case combine_compiler(h_undefproc, g_fstvar):
	case combine_compiler(h_undefproc, g_nxtvar):
	case combine_compiler(h_undefproc, g_0fstvar):
	case combine_compiler(h_undefproc, g_0nxtvar):
	case combine_compiler(h_undefproc, g_1fstvar):
	case combine_compiler(h_undefproc, g_1nxtvar):
	case combine_compiler(h_undefproc, g_2fstvar):
	case combine_compiler(h_undefproc, g_2nxtvar):
	case combine_compiler(h_undefproc, g_3fstvar):
	case combine_compiler(h_undefproc, g_3nxtvar):
	case combine_compiler(c_undefproc, c_call):
	case combine_compiler(c_undefproc, c_lastcall):
compile_undef_proc:
		SDecr(goalptr);
		LDecr(goalptr);
		fprintf(stderr, "Fehler bei CompileUndefProc\n");
		return(ErrorNumber = UndefProcError);

	case combine_compiler(c_goal, c_lastcall):          /*  parameter (long) - headptr */
		/* fall through to combine_compiler(c_4_goal, c_lastcall) */
	case combine_compiler(c_nogoal, c_lastcall):        /*  parameter (long) - headptr */
		/* fall through to combine_compiler(c_4_goal, c_lastcall) */
	case combine_compiler(c_builtin, c_lastcall):       /*  parameter (long) - headptr */
		/* fall through to combine_compiler(c_4_goal, c_lastcall) */
	case combine_compiler(c_cut, c_lastcall):           /*  parameter (long) - headptr */
		/* fall through to combine_compiler(c_4_goal, c_lastcall) */
	case combine_compiler(c_1_goal, c_lastcall):        /*  parameter (long) - headptr */
		/* fall through to combine_compiler(c_4_goal, c_lastcall) */
	case combine_compiler(c_2_goal, c_lastcall):        /*  parameter (long) - headptr */
		/* fall through to combine_compiler(c_4_goal, c_lastcall) */
	case combine_compiler(c_3_goal, c_lastcall):        /*  parameter (long) - headptr */
		/* fall through to combine_compiler(c_4_goal, c_lastcall) */
	case combine_compiler(c_4_goal, c_lastcall):        /*  parameter (long) - headptr */
	case combine_compiler(c_v1_xgoal, c_lastcall):      /*  parameter (long) - goalptr at V1_lastcall return to VAM2P-Code == headptr now */
		SDecr(headptr);
		TraceCodeGeneration("V1_lastcall-Befehl generiert\n");
		gen_lop_code(v1_lastcall, (MW) headptr);
		return (ErrorNumber = Success); /* ??? */

	case combine_compiler(c_xgoal, c_lastcall):          /*  parameter (long) - headptr */
		/* fall through to combine_compiler(c_4_xgoal, c_lastcall) */
	case combine_compiler(c_1_xgoal, c_lastcall):        /*  parameter (long) - headptr */
		/* fall through to combine_compiler(c_4_xgoal, c_lastcall) */
	case combine_compiler(c_2_xgoal, c_lastcall):        /*  parameter (long) - headptr */
		/* fall through to combine_compiler(c_4_xgoal, c_lastcall) */
	case combine_compiler(c_3_xgoal, c_lastcall):        /*  parameter (long) - headptr */
		/* fall through to combine_compiler(c_4_xgoal, c_lastcall) */
	case combine_compiler(c_4_xgoal, c_lastcall):        /* 2 parameters (long) - headptr, pointer to Vam1PCode */
		SDecr(headptr);
		TraceCodeGeneration("V1_lastcall-Befehl generiert\n");
		gen_lop_code(v1_lastcall, (MW) headptr);
		return (ErrorNumber = Success); 

write_xgoal_label:

		goaltemp =  *lastgptr;
		StructDeref(goaltemp, lastgptr);
		CodeStackCheck;
		/* fetch clause-name of last subgoal */
		if (Atom(goaltemp))
			{
			clause_name_lastsubgoal = XtrAtomIndex(goaltemp);
			clause_arity_lastsubgoal = 0;
			}
		else  /* Struct(lastgptr) */
			{
			lastgptr = XtrStruct(goaltemp);
			clause_name_lastsubgoal = FunctIndex(XtrFunctor(lastgptr));
			clause_arity_lastsubgoal = XtrArity(lastgptr);
			}

		flag_v1_xgoal_generated = FALSE;
		if ((clause_name == clause_name_lastsubgoal) && (clause_arity == clause_arity_lastsubgoal))
			{
			/* [headvarcount(short),] v1_procptr (long), headptr (long), v1_codeptr (long), varoffset */
			v1_procptr = find_proc(find_func(clause_name, clause_arity));
			TraceCodeGeneration("V1_xgoal-Befehl generiert\n");
			v1_codeptr[current_goal] = codeptr;

			
			switch (v.head_vars)
				{
				case 1:
				case 2:
				case 3:
				case 4:
				default:
					v1_headptr = headptr; 
					/* [headvarcount(short),] v1_procptr (long), headptr (long), v1_codeptr (long), varoffset */
					gen_soploploplopsop_code(c_v1_xgoal, v.head_vars, (MW) v1_procptr, (MW) v1_headptr, (MW) &v1_codeptr[current_goal], varoffset);
				} /* switch */
			flag_v1_xgoal_generated = TRUE;
			goalbeginptr = codeptr;
			}
		return (ErrorNumber = Success);
	case combine_compiler(c_goal, c_call):                        /*  error */
		fprintf(stderr, "InternalVAM1pCompilerError\n");
		return(ErrorNumber = SystemError);
	}

compile_head_unify:
while (--i >= 0)
	switch(HeadMux(InstrFetch(headptr)))
		{
		case HeadMux(h_const):									/* long */
			LopFetch(m1, headptr);
			gen_lop_code(v1_unify_const, m1);
			break;
		case HeadMux(h_nil):									/* no operand */
			gen_nop_code(v1_unify_nil);
			break;
		case HeadMux(h_list):									/* list - 2 elements */
			if (i == 0)
				gen_nop_code(v1_unify_last_list);
			else
			{
				gen_nop_code(v1_unify_list);
				count_unify_lists++;
			}
			i += 2;
			break;
		case HeadMux(h_struct):									/* structure - functor and n elements */
			m1 = (MW) FunctorFetch(headptr);
			if (i == 0)
				gen_lop_code(v1_unify_last_struct, m1);
			else
			{
				gen_lop_code(v1_unify_struct, m1);
				count_unify_lists++;
			}
			i += ArityFetch(headptr);
			FirstArgIncr(headptr);
			break;
		case HeadMux(h_void):									/* no operand */
			gen_nop_code(v1_unify_void);
			break;
		case HeadMux(h_fstvar):									/* short */
			switch (off1 = SopFetch(headptr))
				{
				case Offset(0):									/* no operand */
					gen_nop_code(v1_unify_0fstvar);   
					break;
				case Offset(1):									/* no operand */
					gen_nop_code(v1_unify_1fstvar);   
					break;
				case Offset(2):									/* no operand */ 
					gen_nop_code(v1_unify_2fstvar);   
					break;
				case Offset(3):									/* no operand */
					gen_nop_code(v1_unify_3fstvar);   
					break;
				default:									/* short */
					gen_sop_code(v1_unify_fstvar, off1);   
					break;
				}
				break;
		case HeadMux(h_nxtvar):									/* short */
			switch (off1 = SopFetch(headptr))
				{
				case Offset(0):									/* no operand */
					gen_nop_code(v1_unify_0nxtvar);
					break;
				case Offset(1):									/* no operand */
					gen_nop_code(v1_unify_1nxtvar);
					break;
				case Offset(2):									/* no operand */
					gen_nop_code(v1_unify_2nxtvar);
					break;
				case Offset(3):									/* no operand */
					gen_nop_code(v1_unify_3nxtvar);
					break;
				default:										/* short */
					gen_sop_code(v1_unify_nxtvar, off1);
					break;
				}
				break;
		case HeadMux(h_fsttemp):										/* short */
			gen_sop_code(v1_unify_fsttmp, SopFetch(headptr));
			break;
		case HeadMux(h_nxttemp):										/* short */
			gen_sop_code(v1_unify_nxttmp, SopFetch(headptr));
			break;
		case HeadMux(h_fsttmp0):										/* no operand */
			gen_nop_code(v1_unify_fsttmp0);
			break;
		case HeadMux(h_nxttmp0):										/* no operand */
			gen_nop_code(v1_unify_nxttmp0); 
			break;
		case HeadMux(h_undefproc):
			return(ErrorNumber = UndefProcError);
		default:
			fprintf(stderr, "CompileHeadUnifyError\n");
			return(ErrorNumber = SystemError);
		}
while (count_unify_lists-- > 0)
	gen_nop_code(v1_pop); 
count_unify_lists = 0;
goto compile;


compile_goal_unify:
while (--i >= 0)
	switch(GoalMux(InstrFetch(goalptr)))
		{
		case GoalMux(g_const):								/* long */
			LopFetch(m1, goalptr);
			gen_lop_code(v1_unify_const, m1);
			break;
		case GoalMux(g_nil):								/* no operand */
			gen_nop_code(v1_unify_nil);
			break;
		case GoalMux(g_list):								/* list - 2 elements */
			if (i == 0)
				gen_nop_code(v1_unify_last_list);
			else
			{
				gen_nop_code(v1_unify_list);
				count_unify_lists++;
			}
			i += 2;
			break;
		case GoalMux(g_struct):								/* structure - functor and  n elements */
			m1 = (MW) FunctorFetch(goalptr);
			if (i == 0)
				gen_lop_code(v1_unify_last_struct, m1);
			else
			{
				gen_lop_code(v1_unify_struct, m1);
				count_unify_lists++;
			}
			i += ArityFetch(goalptr);
			FirstArgIncr(goalptr);
			break;
		case GoalMux(g_void):										/* no operand */
			gen_nop_code(v1_unify_void);
			break;
		case GoalMux(g_fstvar):										/* short */
			gen_sop_code(v1_unify_fstvar, SopFetch(goalptr));
			break;
		case GoalMux(g_nxtvar):										/* short */	
			gen_sop_code(v1_unify_nxtvar, SopFetch(goalptr));
			break;
		case GoalMux(g_0fstvar):									/* no operand */
			gen_nop_code(v1_unify_0fstvar);
			break;
		case GoalMux(g_0nxtvar):									/* no operand */
			gen_nop_code(v1_unify_0nxtvar);
			break;
		case GoalMux(g_1fstvar):									/* no operand */
			gen_nop_code(v1_unify_1fstvar);
			break;
		case GoalMux(g_1nxtvar):									/* no operand */
			gen_nop_code(v1_unify_1nxtvar);
			break;
		case GoalMux(g_2fstvar):									/* no operand */
			gen_nop_code(v1_unify_2fstvar);
			break;
		case GoalMux(g_2nxtvar):									/* no operand */
			gen_nop_code(v1_unify_2nxtvar);
			break;
		case GoalMux(g_3fstvar):									/* no operand */
			gen_nop_code(v1_unify_3fstvar);
			break;
		case GoalMux(g_3nxtvar):									/* no operand */
			gen_nop_code(v1_unify_3nxtvar);
			break;
		default:
			fprintf(stderr, "CompileGoalUnifyError\n");
			return(ErrorNumber = SystemError);
		}
while (count_unify_lists-- > 0)
	gen_nop_code(v1_pop); 
count_unify_lists = 0;
goto compile;

compile_fail:
	goto write_commands_label;	

compile:
	goto write_commands_label;	
}


